## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import pdb
import sys
import logging
import selftest
import importlib
import pathlib
import optparse
import os
import os.path


"""
Runs selftests reporting to stdout.

Usage:
  $ selftest <path or module>

"""

if "SELFTEST_MAIN" not in os.environ:
    # avoid running main twice (during import in a spawned process)
    os.environ["SELFTEST_MAIN"] = "Y"

    def post_mortem(tb, *cmds):
        """for when you use plain assert, it'll throw you in Pdb on failure"""
        p = pdb.Pdb(stdout=sys.stderr)
        p.rcLines.extend(cmds)
        p.reset()
        p.interaction(None, tb)

    def insert_excepthook(new_hook):
        prev_excepthook = sys.excepthook

        def hook(*args):
            prev_excepthook(*new_hook(*args))

        sys.excepthook = hook

    def code_print_excepthook(t, v, tb):
        post_mortem(tb, "list", "exit")
        return t, v, tb

    cwd = pathlib.Path.cwd()
    sys.path.insert(0, cwd.as_posix())

    insert_excepthook(code_print_excepthook)
    insert_excepthook(lambda t, v, tb: (t, v, selftest.utils.filter_traceback(tb)))

    p = optparse.OptionParser(usage="usage: %prog [options] module")
    p.add_option(
        "-f", "--filter", help="only run tests whose qualified name contains FILTER"
    )
    p.add_option(
        "-t",
        "--threshold",
        help="only run tests whose level is >= THRESHOLD",
        choices=[
            l.lower() for l in selftest.levels.levels.keys() if isinstance(l, str)
        ],
    )
    options, args = p.parse_args()

    test_options = {}
    if f := options.filter:
        test_options["filter"] = f
    test_options["threshold"] = options.threshold or "integration"
    selftest.basic_config(**test_options)

    class F(str):
        def format(self, *, levelname="", message="", lineno="", pathname="", **rest):
            levelname = "\033[1mTEST\033[0m" if levelname == "TEST" else levelname
            return f"{levelname}:\033[1m{message}\033[0m:{pathname[-40:]}:{lineno}"

    logging.basicConfig(style="{", format=F("{fake}"))

    root = selftest.get_tester()
    if len(args) == 1:
        p = pathlib.Path(args[0])
        modulename = ".".join(p.parent.parts + (p.stem,))
        logging.getLogger("selftest").log(
            selftest.tester.default_loglevel, f"importing {modulename}"
        )
        importlib.import_module(modulename)
        root.log_stats()
    else:
        p.print_usage()



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import multiprocessing
import asyncio


def spawn(f):
    ctx = multiprocessing.get_context("spawn")
    p = ctx.Process(target=f)
    p.start()
    return p


def fork(f):
    ctx = multiprocessing.get_context("fork")
    p = ctx.Process(target=f)
    p.start()
    return p


def subprocess():
    """when spawned, imports selftest freshly, potentially running al tests
        but this is prevented by default, use subprocess=True when needed
    when forked, use existing loaded selftest, not running tests
        so no special measures are taken and both tests run
    """
    import selftest

    test = selftest.get_tester("subprocess")

    @test
    def one_test_not_run():
        print("In a forked child, I run.")

    @test(subprocess=True)
    def one_does_run():
        print("In any child, I run.")


def integration_test(test):
    @test
    def import_submodule(stdout, stderr):
        from selftest.tests.sub_module_ok import marker

        test.eq("TESTER: <Tester 'selftest.tests.sub_module_ok' created at:\n"
                '  File "<string>", line 1, in <module>\n>\n', stdout.getvalue())
        test.eq("UNIT:selftest.tests.sub_module_ok.test_one\n", stderr.getvalue())

    try:
        with test.stdout as s, test.stderr as r:

            @test
            def import_submodule_failure():
                from .tests import sub_module_fail

        test.fail("Should have failed.")
    except AssertionError as e:
        test.eq("fail I will", str(e))
        m = r.getvalue()
        test.eq(
            m,
            "INTEGRATION:selftest.integrationtests.integration_test.import_submodule_failure\nUNIT:selftest.tests.sub_module_fail.tiedeldom\n",
        )

    with test.raises(AssertionError, "Use combine:3 instead of combine=3"):

        @test
        def fixture_args_as_annotain_iso_defaul(combine=3):
            """fixture args are not default args: '=' instead of ':' raises error"""
            pass

    @test(bind=True)
    def by_default_do_not_run_in_spawned_processes(stdout, stderr):
        spawn(subprocess).join()
        test.contains(stderr.getvalue(), "subprocess.one_does_run")
        test.eq("In any child, I run.\n", stdout.getvalue())

    @test(bind=True)
    def test_do_not_run_in_child_processes(stdout):
        fork(subprocess).join()
        test.eq("In a forked child, I run.\nIn any child, I run.\n", stdout.getvalue())

    @test.fixture
    def async_combine(a):
        yield a

    with test.raises(AssertionError, "Use async_combine:3 instead of async_combine=3"):

        @test
        async def fixture_args_as_annotain_iso_defaul(async_combine=3):
            """fixture args are not default args: '=' instead of ':' raises error"""
            pass

    @test
    def combine_test_with_options_and_test_arg():
        trace = []
        with test.child(keep=True, my_opt=42) as tst:

            @tst
            def f0(test):
                trace.append(test.option_get("my_opt"))

            def f1(test):
                trace.append(test.option_get("my_opt"))

            # at this point, f0 and f1 are equivalent; @tst does not modify f0
            # so when run, a new test context must be provided:
            test(f0)
            test(f0, my_opt=76)
            test(f0, f1, my_opt=93)
            assert [42, None, 76, 93, 93] == trace, trace

    class async_fixtures:
        """here because async fixtures need async tests"""

        @test.fixture
        async def my_async_fixture():
            await asyncio.sleep(0)
            yield "async-42"

        @test
        async def async_fixture_with_async_with():
            async with test.my_async_fixture as f:
                assert "async-42" == f
            try:
                with test.my_async_fixture:
                    assert False
                test.fail()
            except Exception as e:
                test.startswith(str(e), "Use 'async with' for ")

        @test
        async def async_fixture_as_arg(my_async_fixture):
            test.eq("async-42", my_async_fixture)

        try:

            @test
            def only_async_funcs_can_have_async_fixtures(my_async_fixture):
                assert False, "not possible"

        except AssertionError as e:
            test.eq(
                f"function 'only_async_funcs_can_have_async_fixtures' cannot have async fixture 'my_async_fixture'.",
                str(e),
            )

        @test.fixture
        async def with_nested_async_fixture(my_async_fixture):
            yield f">>>{my_async_fixture}<<<"

        @test
        async def async_test_with_nested_async_fixtures(with_nested_async_fixture):
            test.eq(">>>async-42<<<", with_nested_async_fixture)

        @test
        async def mix_async_and_sync_fixtures(fixture_C, with_nested_async_fixture):
            test.eq(">>>async-42<<<", with_nested_async_fixture)
            test.eq(252, fixture_C)

        @test.fixture
        def area(r, d=1):
            import math

            yield round(math.pi * r * r, d)

        @test
        async def fixtures_with_2_args_async(area: (3, 2)):
            test.eq(28.27, area)

    class tmp_files:
        path = [None]

        @test
        def temp_sync(tmp_path):
            assert tmp_path.exists()

        @test(bind=True)
        def temp_file_removal(tmp_path):
            path[0] = tmp_path / "aap"
            path[0].write_text("hello")

        @test(bind=True)
        def temp_file_gone():
            assert not path[0].exists()

        @test(bind=True)
        async def temp_async(tmp_path):
            assert tmp_path.exists()

        @test
        def temp_dir_with_file(tmp_path: "aap"):
            assert str(tmp_path).endswith("/aap")
            tmp_path.write_text("hi monkey")
            assert tmp_path.exists()



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import operator  # operators for asserting
import builtins  # operators for asserting
import inspect  # operators for asserting


class _Operators:
    def __call__(self, tester, f):
        return f

    def lookup(self, runner, name, truth=bool):
        if name in ("comp", "complement"):

            class Not:
                def __getattr__(inner, name):
                    return self.lookup(runner, name, truth=operator.not_)

            return Not()

        def call_operator(*oargs, diff=None):
            args = oargs  # TODO test
            SELFTEST_INTERNAL = 1
            if hasattr(operator, name):
                op = getattr(operator, name)
            elif hasattr(builtins, name):
                op = getattr(builtins, name)
            elif hasattr(inspect, name):
                op = getattr(inspect, name)
            else:
                op = getattr(args[0], name)
                args = oargs[1:]
            if truth(op(*args)):
                return True
            if diff:
                raise AssertionError(diff(*oargs))
            else:
                raise AssertionError(op.__name__, *oargs)

        return call_operator


operators_hook = _Operators()


def operators_test(self_test):
    @self_test
    def test_isinstance():
        self_test.isinstance(1, int)
        self_test.isinstance(1.1, (int, float))
        try:
            self_test.isinstance(1, str)
            self.fail()
        except AssertionError as e:
            assert str(e) == "('isinstance', 1, <class 'str'>)"
        try:
            self_test.isinstance(1, (str, dict))
            self.fail()
        except AssertionError as e:
            assert str(e) == "('isinstance', 1, (<class 'str'>, <class 'dict'>))"

    @self_test
    def use_builtin():
        """you could use any builtin as well; there are not that much useful options in module builtins though
        we could change the priority of lookups, now it is: operator, builtins, <arg[0]>. Maybe reverse?
        """
        self_test.all([1, 2, 3])
        try:
            self_test.all([False, 2, 3])
            self.fail()
        except AssertionError as e:
            assert str(e) == "('all', [False, 2, 3])", e

        class A:
            pass

        class B(A):
            pass

        self_test.issubclass(B, A)
        self_test.hasattr([], "append")

    @self_test(keep=True)
    def use_inspect():
        def f():
            pass

        self_test.isfunction(f)



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

""" Defines the Tester

    NB: a stand alone Tester (self_test) is created as to not interfere with
        the system wide default runner, which may not exists as long as we
        are bootstrapping.

    NB: Tester is tested using Tester itself. Tester is bootstrapped in steps,i
        starting with a simple Tester of which the capabilities are gradually
        extended.

    TODO nav gebruik in metastreams-server:
    2. Make async tests reuse event loop if available

"""

import itertools # iterate over stack
import traceback # capture creation of Tester objects
import inspect  # for recognizing functions, generators etc
import contextlib  # child as context
import collections  # chain maps for hierarchical Tester and Counter
import logging  # output to logger
import os  # controlling env
import sys  # maxsize
import io  # formatting tests


logger = logging.getLogger("selftest")
# in order to have tests shown when logging is unconfigured, we log at a level slightly
# higher than the default level WARNING (30)
default_loglevel = 33
logging.addLevelName(default_loglevel, "TEST")  # TODO make configurable


""" by default, tests are suppressed in subprocesses because many tests run
    during import, which can easily lead to an endless loop.
    Use subprocess=True when needed.
    This is NOT the same as preventing __main__ from running multiple times!
"""
is_subprocess = "SELFTEST_PARENT" in os.environ
os.environ["SELFTEST_PARENT"] = "Y"


defaults = dict(
    keep=False,  # Ditch test functions after running
    run=True,  # Run test immediately
    subprocess=False,  # Do not run test when in a subprocess
)


class Tester:
    """Main tool for running tests across modules and programs."""

    def __call__(self, *functions, **options):
        """Runs tests, with options; useful as decorator."""
        SELFTEST_INTERNAL = 1
        if functions and options:
            return self(**options)(*functions)
        if options:
            return self.getChild(**options)
        for f in functions:
            result = self._run(f)
        return result

    def getChild(self, *_, **__):
        return Tester(*_, parent=self, **__)

    get_child = getChild

    @contextlib.contextmanager
    def child(self, *_, **__):
        yield self.getChild(*_, **__)

    def fail(self, *args, **kwds):
        raise AssertionError(*args, *(kwds,) if kwds else ())

    def __init__(self, name=None, parent=None, **options):
        """Not to be instantiated directly, use selftest.getTester() instead"""
        fullstack = traceback.extract_stack().format()
        stack = (f for f in fullstack if '/selftest/' not in f and '_bootstrap' not in f)
        self._stack = ''.join(stack)
        self._parent = parent
        if parent:
            self._name = (
                parent._name + "." + name
                if name and parent._name
                else name or parent._name
            )
            self._options = parent._options.new_child(m=options)
        else:
            self._name = name
            self._options = collections.ChainMap(options, defaults)
        self.stats = collections.Counter(found=0, run=0)
        self._loghandlers = []
        self._keep_path = sys.path.copy()
        self._keep_meta = sys.meta_path.copy()
        self._keep_mods = sys.modules.copy()

    def option_get(self, name, default=None):
        return self._options.get(name, default)

    def option_setdefault(self, name, default):
        return self._options.maps[0].setdefault(name, default)

    def option_enumerate(self, name):
        """enumerate options from parents, flattening sequences"""
        for m in self._options.maps:
            value = m.get(name)
            if isinstance(value, (list, tuple)):
                yield from reversed(value)
            elif value is not None:
                yield value

    def _stat(self, key):
        self.stats[key] += 1
        if p := self._parent:
            p._stat(key)

    def all_hooks(self):
        return self.option_enumerate("hooks")

    def _logtest(self, f, msg):
        loggername = f"{logger.name}.{self._name}" if self._name else logger.name
        record = logging.LogRecord(
            loggername,  # name of logger
            self.option_get(
                "loglevel", default_loglevel
            ),  # level under which test messages get logged
            f.__code__.co_filename if f else None,  # source file where test is
            f.__code__.co_firstlineno if f else None,  # line where test is
            msg,  # message
            (),  # args (passed to message.format)
            None,  # exc_info
            f.__qualname__ if f else None,  # name of the function invoking test.<op>
            None,  # text representation of stack
        )
        for hook in self.all_hooks():
            try:
                record = hook.logrecord(self, f, record)
            except AttributeError:
                pass
        logger.handle(record)

    def _run(self, test_func, *app_args, **app_kwds):
        """Runs hooks and and runs result."""
        SELFTEST_INTERNAL = 1
        self._stat("found")
        orig_test_func = test_func
        if not is_subprocess or self.option_get("subprocess", False):
            for hook in self.all_hooks():
                if not (test_func := hook(self, test_func)):
                    break
            else:
                if self.option_get("run"):
                    self._stat("run")
                    self._logtest(
                        orig_test_func, orig_test_func.__qualname__
                    )  # TODO pass test func?
                    test_func(*app_args, **app_kwds)
        return orig_test_func if self.option_get("keep") else None

    def log_stats(self):
        name = self._name + ".stats: " if self._name else "stats: "
        logger.log(
            self.option_get("loglevel", default_loglevel),
            name + ", ".join(f"{k}: {v}" for k, v in self.stats.most_common()),
        )

    def __getattr__(self, name):
        for hook in self.all_hooks():
            try:
                return hook.lookup(self, name)
            except AttributeError:
                pass

    def __str__(self):
        CR = '\n'
        return f"<Tester {self._name!r} created at:{CR}{self._stack}>"


from numbers import Number


class any_number:
    def __init__(self, lo, hi):
        self._lo = lo
        self._hi = hi

    def __eq__(self, rhs):
        return isinstance(rhs, Number) and self._lo <= rhs <= self._hi


from .operators import operators_hook
from .levels import levels_hook

from sys import argv

if "selftest.selftest" in argv:
    argv.remove("selftest.selftest")
    self_test = Tester(
        "selftest-self-tests", hooks=(operators_hook, levels_hook), level="unit"
    )  # separate runner for bootstrapping/self testing
else:

    class Ignore:
        def __call__(self, runner, func):
            pass

        def lookup(self, runner, name):
            return self

    def get_tester():
        """ marker function """
        return Tester(hooks=[Ignore()])
    self_test = get_tester()
    # self_test.addHandler(logging.NullHandler())


@self_test
def example_test():
    assert any_number(1, 10) == 1
    assert any_number(1, 10) == 2
    assert any_number(1, 10) == 5
    assert any_number(1, 10) == 9
    assert any_number(1, 10) == 10
    assert not any_number(9, 15) == 1
    assert not any_number(9, 15) == 8
    assert not any_number(9, 15) == 16
    assert not any_number(9, 15) == 33
    assert not any_number(9, 15) == None
    assert not any_number(9, 15) == "A"
    assert not any_number(9, 15) == object


@self_test
def new_assert():
    assert self_test.eq(1, 1)
    assert self_test.lt(1, 2)  # etc

    try:
        self_test.eq(1, 2)
        self_test.fail()
    except AssertionError as e:
        assert ("eq", 1, 2) == e.args, e.args

    try:
        self_test.ne(1, 1)
        self_test.fail()
    except AssertionError as e:
        assert "('ne', 1, 1)" == str(e), str(e)


@self_test
def test_fail():
    try:
        self_test.fail("plz fail")
        raise AssertionError("FAILED to fail")
    except AssertionError as e:
        assert "plz fail" == str(e), e

    try:
        self_test.fail("plz fail", info="more")
        raise AssertionError("FAILED to fail")
    except AssertionError as e:
        assert "('plz fail', {'info': 'more'})" == str(e), e


@self_test
def not_operator():
    self_test.comp.contains("abc", "d")
    try:
        self_test.comp.contains("abc", "b")
        self_test.fail()
    except AssertionError as e:
        assert "('contains', 'abc', 'b')" == str(e), e


@self_test
def call_method_as_operator():
    self_test.endswith("aap", "ap")


# self_test.<op> is really just assert++ and does not need @test to run 'in'
self_test.eq(1, 1)

try:
    self_test.lt(1, 1)
except AssertionError as e:
    assert "('lt', 1, 1)" == str(e), str(e)


@self_test
def test_testop_has_args():
    try:

        @self_test
        def nested_test_with_testop():
            x = 42
            y = 67
            self_test.eq(x, y)
            return True

    except AssertionError as e:
        assert hasattr(e, "args")
        assert 3 == len(e.args)
        self_test.eq("('eq', 42, 67)", str(e))


@self_test
def test_calls_other_test():
    @self_test(keep=True)
    def test_a():
        assert 1 == 1
        return True

    @self_test.child()
    def test_b():
        assert test_a()


@self_test
def call_test_with_complete_context():
    @self_test(keep=True)
    def a_test():
        assert True

    assert a_test, a_test
    self_test(a_test)


@contextlib.contextmanager
def intercept():
    records = []

    class intercept:
        level = 33
        handle = records.append

    root_logger = logging.getLogger()
    root_logger.addHandler(intercept)
    try:
        yield records
    finally:
        root_logger.removeHandler(intercept)


@contextlib.contextmanager
def configure_stream_hander():
    selftest_logger = logging.getLogger("selftest")
    s = io.StringIO()
    handler = logging.StreamHandler(s)
    handler.setFormatter(logging.Formatter(fmt=logging._STYLES["%"][1]))
    selftest_logger.addHandler(handler)
    try:
        yield s
    finally:
        selftest_logger.removeHandler(handler)


class logging_handlers:
    @self_test
    def tester_delegates_to_root_logger():
        with intercept() as i:
            tester = Tester("free")

            @tester
            def my_output_goes_to_root_logger():
                assert 1 == 1  # hooks for operators not present

            self_test.eq(
                "logging_handlers.tester_delegates_to_root_logger.<locals>.my_output_goes_to_root_logger",
                i[0].funcName,
            )

    @self_test
    def logrecords_contents():
        tester = Tester("unconfigured_logging", hooks=[levels_hook])
        with intercept() as records:
            _line_ = inspect.currentframe().f_lineno

            @tester
            def a_test():
                assert 1 == 1

            self_test.eq((), records[0].args)
            self_test.lt(1673605231.8793523, records[0].created)
            self_test.eq(None, records[0].exc_info)
            self_test.eq("tester.py", records[0].filename)
            self_test.eq(
                "logging_handlers.logrecords_contents.<locals>.a_test",
                records[0].funcName,
            )
            self_test.eq("TEST", records[0].levelname)
            self_test.eq(33, records[0].levelno)
            self_test.eq(_line_ + 2, records[0].lineno)
            self_test.eq("tester", records[0].module)
            self_test.truth(10 < records[0].msecs < 1000)
            self_test.eq(
                "UNIT:logging_handlers.logrecords_contents.<locals>.a_test",
                records[0].msg,
            )
            self_test.eq("selftest.unconfigured_logging", records[0].name)
            self_test.endswith(records[0].pathname, "selftest/selftest/tester.py")
            self_test.truth(2 < records[0].process < 100000)
            self_test.eq("MainProcess", records[0].processName)
            self_test.truth(1 < records[0].relativeCreated < 1000)
            self_test.eq(None, records[0].stack_info)
            self_test.truth(1 < records[0].thread < sys.maxsize)
            self_test.eq("MainThread", records[0].threadName)
            self_test.eq(40, records[0].testlevel)
            self_test.eq("UNIT", records[0].testlevelname)

    @self_test
    def tester_with_default_formatting():
        with configure_stream_hander() as s:
            tester = Tester("default_formatting", hooks=[levels_hook])

            @tester
            def a_test():
                assert 1 == 1  # hooks for operators not present

            self_test.eq(
                "TEST:selftest.default_formatting:UNIT:logging_handlers.tester_with_default_formatting.<locals>.a_test\n",
                s.getvalue(),
            )

    @self_test
    def tester_without_a_name():
        with configure_stream_hander() as s:
            tester = Tester(hooks=[levels_hook])

            @tester
            def a_test():
                assert 1 == 1  # hooks for operators not present

            self_test.eq(
                "TEST:selftest:UNIT:logging_handlers.tester_without_a_name.<locals>.a_test\n",
                s.getvalue(),
            )

    @self_test
    def sub_tester():
        with configure_stream_hander() as s:
            main = Tester("main", hooks=[levels_hook])
            sub1 = main.getChild("sub1")
            sub2 = sub1.getChild("sub2")

            @sub1
            def sub1_test():
                pass

            @main
            def main_test():
                pass

            @sub2
            def sub2_test():
                pass

            loglines = s.getvalue().splitlines()
            self_test.eq(
                "TEST:selftest.main.sub1:UNIT:logging_handlers.sub_tester.<locals>.sub1_test",
                loglines[0],
            )
            self_test.eq(
                "TEST:selftest.main:UNIT:logging_handlers.sub_tester.<locals>.main_test",
                loglines[1],
            )
            self_test.eq(
                "TEST:selftest.main.sub1.sub2:UNIT:logging_handlers.sub_tester.<locals>.sub2_test",
                loglines[2],
            )

    @self_test
    def tester_with_handler_failing():
        with configure_stream_hander() as s:
            main = Tester("main", hooks=[levels_hook])
            try:

                @main
                def a_failing_test():
                    assert 1 == 2

            except AssertionError:
                pass
            self_test.eq(
                "TEST:selftest.main:UNIT:logging_handlers.tester_with_handler_failing.<locals>.a_failing_test\n",
                s.getvalue(),
            )

    @self_test
    def run_or_not():
        @self_test(run=False)
        def not_run():
            assert "do not" == "run me"



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import difflib  # show diffs on failed tests with two args
import pprint  # show diffs on failed tests with two args
import functools
import inspect
from .prrint import prrint, format as forrmat


def diff(a, b, format=pprint.pformat):
    """Produces diff of textual representation of a and b."""
    return "\n" + "\n".join(
        difflib.ndiff(format(a).splitlines(), format(b).splitlines())
    )


def diff2(a, b):
    """diff based on own prrint; suitable for PODs"""
    return diff(a, b, format=forrmat)


class DiffHook:
    def __call__(self, runner, func):
        return func

    def lookup(self, runner, name):
        if name == "diff":
            return diff
        if name == "diff2":
            return diff2
        if name == "prrint":
            return prrint
        raise AttributeError


diff_hook = DiffHook()


def diff_test(self_test):
    self_test = self_test.getChild(hooks=[diff_hook])

    @self_test
    def idea_for_dumb_diffs():
        # if you want a so called smart diff: the second arg of assert is meant for it.
        # Runner supplies a generic diff between two pretty printed values
        a = [7, 1, 2, 8, 3, 4]
        b = [1, 2, 9, 3, 4, 6]
        d = self_test.diff(a, b)
        assert str == type(str(d))

        try:
            assert a == b, self_test.diff(a, b)
        except AssertionError as e:
            assert """
- [7, 1, 2, 8, 3, 4]
?  ---      ^

+ [1, 2, 9, 3, 4, 6]
?        ^      +++
""" == str(
                e
            )

        try:
            self_test.eq(a, b, diff=self_test.diff)
        except AssertionError as e:
            assert """
- [7, 1, 2, 8, 3, 4]
?  ---      ^

+ [1, 2, 9, 3, 4, 6]
?        ^      +++
""" == str(
                e
            ), e

        # use of a function from elswhere
        a = set([7, 1, 2, 8, 3, 4])
        b = set([1, 2, 9, 3, 4, 6])
        try:
            assert a == b, set.symmetric_difference(a, b)
        except AssertionError as e:
            assert "{6, 7, 8, 9}" == str(e), e
        try:
            self_test.eq(a, b, diff=set.symmetric_difference)
        except AssertionError as e:
            assert "{6, 7, 8, 9}" == str(e), e

    @self_test
    def diff2_sorting_including_uncomparables():
        msg = """
  {
+   1,
-   <class 'dict'>:
?           ^^^   ^

+   <class 'str'>,
?           ^ +  ^

-     <class 'bool'>,
  }"""
        try:
            self_test.eq({dict: bool}, {str, 1}, diff=self_test.diff2)
        except AssertionError as e:
            assert msg == str(e), e



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##


def fullname(func):
    return f"{func.__module__}.{func.__qualname__}".replace(".<locals>", "")


class FilterHook:
    def __call__(self, runner, func):
        f = runner.option_get("filter", "")
        if f in fullname(func):
            return func

    def logrecord(self, runner, func, record):
        record.msg = fullname(func)
        return record


filter_hook = FilterHook()


def filter_test(self_test):
    my_test = self_test(hooks=[filter_hook])

    @my_test
    def hook_but_no_filter_given():
        pass

    @my_test
    def aap_mies():
        with my_test.child(filter="noot") as with_noot:
            r = [0, 0, 0]

            class aap:
                @with_noot
                def aap():
                    r[0] = 1

                @with_noot
                def noot():
                    r[1] = 1

                    @with_noot
                    def mies():
                        r[2] = 1

            assert r == [0, 1, 1], r
        assert {"found": 3, "run": 2} == with_noot.stats, with_noot.stats

    @my_test
    def remove_locals():
        r = [0]

        @my_test(filter="remove_locals.vuur")
        def vuur():
            r[0] = 1

        assert r == [1], r

    @my_test
    def filter_on_module():
        r = [0]

        @my_test(filter="selftest.filter.filter_test.filter_on_module.boom")
        def boom():
            r[0] = 1

        assert r == [1], r



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

"""
    This pretty printer aims to make diffs of plain old python (POD) objects more usable.
    It does so by:
      1. making sure every atomic value (str, int, etc) is on a separate line
      2. markup symbols like (, { and : do not interfere by puttin them on a separate line
      3. long atomic values (like str) are continuous; the terminal will help you.
      4. use repr for any unsupported value
"""

import sys
from io import StringIO
from dataclasses import dataclass


def _format(data, indent, seen=None, sep=","):
    write = indent.write
    if id(data) in seen and isinstance(data, (list, dict, tuple, set)):
        write("...", sep)
    else:
        seen.add(id(data))
        if isinstance(data, list):
            if data:
                write("[")
                for v in data:
                    _format(v, indent(), seen)
                write("]", sep)
            else:
                write("[]", sep)
        elif isinstance(data, dict):
            if data:
                write("{")
                for k, v in sorted(data.items(), key=lambda item: str(item[0])):
                    _format(k, indent(), seen, sep=":")
                    _format(v, indent()(), seen)
                write("}", sep)
            else:
                write("{}", sep)
        elif isinstance(data, tuple):
            if data:
                write("(")
                for v in data:
                    _format(v, indent(), seen)
                write(")", sep)
            else:
                write("()", sep)
        elif isinstance(data, set):
            if data:
                write("{")
                for v in sorted(data, key=str):
                    _format(v, indent(), seen)
                write("}", sep)
            else:
                write("set()", sep)
        else:
            write(repr(data), sep)


@dataclass
class Indenter:
    out: StringIO
    i: int = 0
    INDENT: int = 2

    def __call__(self):
        return Indenter(self.out, self.i + self.INDENT)

    def write(self, *data):
        write = self.out.write
        write(self.i * " ")
        for d in data:
            write(d)
        write("\n")


def format(data):
    out = StringIO()
    _format(data, Indenter(out), seen=set(), sep="")
    return out.getvalue()


def prrint(data):
    _format(data, Indenter(sys.stdout), seen=set(), sep="")


def prrint_test(test):
    from .fixtures import fixtures_hook, stdout

    test = test.getChild(hooks=[fixtures_hook], fixtures={"stdout": stdout})

    @test
    def empty():
        test.eq("[]\n", format([]))
        test.eq("{}\n", format({}))
        test.eq("''\n", format(""))
        test.eq("()\n", format(()))
        test.eq("set()\n", format(set()))

    @test
    def one_element():
        test.eq("[\n  42,\n]\n", format([42]))
        test.eq(
            """{
  1:
    'a',
}
""",
            format({1: "a"}),
        )
        test.eq("''\n", format(""))
        test.eq("(\n  42,\n)\n", format((42,)))
        test.eq("[\n  42,\n]\n", format([42]))
        test.eq("{\n  42,\n}\n", format({42}))

    @test
    def separators():
        x = format([1, (2,), {3: 3}, {4}, [5]])
        test.eq(
            """[
  1,
  (
    2,
  ),
  {
    3:
      3,
  },
  {
    4,
  },
  [
    5,
  ],
]
""",
            x,
        )

    @test
    def recursion():
        a = {1: None}
        a[1] = a
        x = format(a)
        test.eq(
            """{
  1:
    ...,
}
""",
            x,
        )

    @test
    def set_sorted():
        x = format({"noot", "mies", "aap"})
        test.eq(
            """{
  'aap',
  'mies',
  'noot',
}
""",
            x,
        )

    @test
    def set_sorted_uncomparables():
        x = format({str, dict, bool})
        test.eq("{\n  <class 'bool'>,\n  <class 'dict'>,\n  <class 'str'>,\n}\n", x)

    @test
    def dict_sorted():
        x = format({"noot": 3, "mies": 2, "aap": 1})
        test.eq(
            """{
  'aap':
    1,
  'mies':
    2,
  'noot':
    3,
}
""",
            x,
        )

    @test
    def dict_sorted_uncomparables():
        x = format({str: 2, dict: 1, bool: 3})
        test.eq(
            "{\n  <class 'bool'>:\n    3,\n  <class 'dict'>:\n    1,\n  <class 'str'>:\n    2,\n}\n",
            x,
        )

    @test
    def beetje_echt():
        x = format(
            {
                "pred_a": [
                    {
                        "pred_b": [
                            {"@value": "32", "@type": "number"},
                            {"@value": "more", "@type": "text"},
                        ]
                    }
                ]
            }
        )
        test.eq(
            """{
  'pred_a':
    [
      {
        'pred_b':
          [
            {
              '@type':
                'number',
              '@value':
                '32',
            },
            {
              '@type':
                'text',
              '@value':
                'more',
            },
          ],
      },
    ],
}
""",
            x,
        )

    @test
    def allez():
        x = format({("aa", "bb"): {("cc", ("dd",)): ("ee",)}, ("ff",): ()})
        test.eq(
            """{
  (
    'aa',
    'bb',
  ):
    {
      (
        'cc',
        (
          'dd',
        ),
      ):
        (
          'ee',
        ),
    },
  (
    'ff',
  ):
    (),
}
""",
            x,
        )

    @test
    def very_long_string_uninterrupted():
        long_string = "aapnootmies" * 100
        x = format((long_string,))
        test.eq(
            f"""(
  '{long_string}',
)
""",
            x,
        )

    @test
    def via_stdout(stdout):
        d = {"a": ("b", 42)}
        f = format(d)
        prrint(d)
        test.eq(f, stdout.getvalue())

    def nndiff(a, b):
        import difflib
        import re

        r = re.compile("(\s*)(\S+)(\s*)")
        al = [r.fullmatch(l).group(1, 2, 3) for l in a]
        bl = [r.fullmatch(l).group(1, 2, 3) for l in b]
        dl = difflib.ndiff([l[1] for l in al], [l[1] for l in bl])
        return dl

    @test
    def diff2_whitespace():
        a = [" a "]
        b = ["  a "]
        d = nndiff(a, b)
        test.eq("  a", "".join(d))



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import inspect
import asyncio
import concurrent


# NB: this supports up to 10 levels of nested async tests
thread = concurrent.futures.ThreadPoolExecutor(max_workers=10).submit


""" Defines a hook for supporting async functions as test.
    It does so by executing the coroutine in an event loop in
    a separate thread. """


def async_hook(runner, func):
    def may_be_async(*a, **k):
        SELFTEST_INTERNAL = 1
        coro_or_result = func(*a, **k)
        if inspect.iscoroutine(coro_or_result):

            async def with_options():
                SELFTEST_INTERNAL = 1
                try:
                    if scbd := runner.option_get("slow_callback_duration"):
                        asyncio.get_running_loop().slow_callback_duration = scbd
                    await asyncio.wait_for(
                        asyncio.shield(coro_or_result), runner.option_get("timeout", 1)
                    )
                except asyncio.TimeoutError as e:
                    raise TimeoutError(func.__name__) from None

            try:
                asyncio.get_running_loop()
                loop = True
            except RuntimeError as e:
                loop = False  # avoid this exception to be reported
            if not loop:
                asyncio.run(with_options(), debug=runner.option_get("debug", True))
            else:
                thread(
                    asyncio.run, with_options(), debug=runner.option_get("debug", True)
                ).result()
        else:
            return coro_or_result

    return may_be_async


def async_test(self_test):
    import threading

    main = threading.current_thread()
    with self_test.child(hooks=(async_hook,)) as atest:
        atest.eq({"found": 0, "run": 0}, atest.stats)
        loops = [None, None, None, None]
        threads = [None, None, None, None]

        @atest
        async def an_async_test():
            loops[0] = asyncio.get_running_loop()
            threads[0] = threading.current_thread()

        @atest
        async def parent():
            loops[1] = asyncio.get_running_loop()
            threads[1] = threading.current_thread()

            @atest
            async def nested_async_test():
                loops[2] = asyncio.get_running_loop()
                threads[2] = threading.current_thread()

            @atest
            def nested_sync_test():
                loops[3] = asyncio.get_running_loop()
                threads[3] = threading.current_thread()

            assert threads[0] == main
            assert threads[1] == main
            assert threads[2] != main
            assert threads[3] == main
            assert len(loops) == 4
            assert loops[0]
            assert loops[1]
            assert loops[1] != loops[0]
            assert loops[2]
            assert loops[2] != loops[0]  # each async test has own loop
            assert loops[2] != loops[1]
            assert loops[3] == loops[1]  # still same loop

        try:

            @atest
            async def with_failures():
                assert 1 == 2, "1 is not 2"
                raise Exception

        except AssertionError as e:
            assert str(e) == "1 is not 2", e

        try:

            @atest(timeout=0.05)
            async def async_timeout_raises_standard_TimeoutError():
                await asyncio.sleep(0.10)
                assert False, "should have raised timeout"

        except TimeoutError as e:
            assert "async_timeout_raises_standard_TimeoutError" == str(e)

        import sys, io, time

        try:
            sys.stderr = io.StringIO()
            sys.stdout = io.StringIO()

            @atest
            async def debug_warnings_on_by_default():
                asyncio.get_running_loop().call_soon(time.sleep, 0.11)

            e = sys.stderr.getvalue()
            s = sys.stdout.getvalue()

            @atest
            def asserts():
                assert "" == s, s
                assert "Executing <Handle sleep(0.11) created at" in e, e
                assert "took 0.11" in e, e
                assert " seconds" in e, e

        finally:
            sys.stderr = sys.__stderr__
            sys.stdout = sys.__stdout__

        try:
            sys.stderr = io.StringIO()

            @atest(keep=True)
            async def default_slow_callback_duration():
                assert asyncio.get_running_loop().slow_callback_duration == 0.1

            @atest(slow_callback_duration=0.12)
            async def slow_callback_duration_option():
                asyncio.get_running_loop().call_soon(time.sleep, 0.11)

            atest(default_slow_callback_duration)
            s = sys.stderr.getvalue()
            assert "Executing <Handle sleep(0.11) created at" not in s, s
            assert "took 0.110 seconds" not in s, s
        finally:
            sys.stderr = sys.__stderr__

        @atest
        def assert_stats():
            atest.eq({"found": 12, "run": 12}, atest.stats)



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import tempfile  # for tmp_path
import pathlib  # for tmp_path
import inspect  # detect generators (for context manager)
import asyncio  # access to loop
import contextlib  # exit stacks
import asyncio  # support for async test and fixtures
import sys  # fiddling with stdout/err
import os  # fiddling with stdout/err file descriptors
import io  # StringIO for testing


from .utils import asyncio_filtering_exception_handler, ensure_async_generator_func
from .utils import extend_closure
from .utils import ArgsCollectingContextManager, ArgsCollectingAsyncContextManager


__all__ = ["std_fixtures"]


def get_fixture(runner, name):
    """synthetic fixture 'test'"""
    if name == "test":

        def test():
            yield runner

        return test
    """ find fixture in hierarchical maps """
    for map in runner.option_enumerate("fixtures"):
        if name in map:
            return map[name]


def add_fixture(runner, func):
    """add fixture to leaf map"""
    runner.option_setdefault("fixtures", {})[func.__name__] = func


class _Fixtures:
    """Activates all fixtures recursively, then runs the test function."""

    def __init__(self, runner, func):
        self.runner = runner
        self.func = func
        self.__name__ = (
            func.__name__
        )  # support getting name for exceptions etc #TODO test

    @classmethod
    def lookup(clz, tester, name):
        """supports looking up fixtues as test.<fixture> and setting them with @test.fixture"""
        if name == "fixture":

            def fixture(func):
                assert inspect.isgeneratorfunction(func) or inspect.isasyncgenfunction(
                    func
                ), func
                bound_f = extend_closure(func)
                add_fixture(tester, bound_f)
                return bound_f

            return fixture
        elif fx := get_fixture(tester, name):
            fx_bound = _Fixtures(tester, fx)
            if inspect.isgeneratorfunction(fx):
                return ArgsCollectingContextManager(fx_bound)
            if inspect.isasyncgenfunction(fx):
                return ArgsCollectingAsyncContextManager(fx_bound)
            raise ValueError(f"not an (async) generator: {fx}")
        raise AttributeError

    def __call__(self, *args, **kwds):
        SELFTEST_INTERNAL = 1
        if inspect.iscoroutinefunction(self.func):
            return self.async_run_with_fixtures(*args, **kwds)
        else:
            with contextlib.ExitStack() as contextmgrstack:
                return self.run_recursively(self.func, contextmgrstack, *args, **kwds)

    def get_fixtures_except_for(self, f, except_for):
        """Finds all fixtures, skips those in except_for (overridden fixtures)"""
        SELFTEST_INTERNAL = 1

        def args(p):
            a = () if p.annotation == inspect.Parameter.empty else p.annotation
            assert (
                p.default == inspect.Parameter.empty
            ), f"Use {p.name}:{p.default} instead of {p.name}={p.default}"
            return a if isinstance(a, tuple) else (a,)

        return [
            (get_fixture(self.runner, name), args(p))
            for name, p in inspect.signature(f).parameters.items()
            if get_fixture(self.runner, name) and name not in except_for
        ]

    # vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv

    def run_recursively(self, f, contextmgrstack, *args, **kwds):
        SELFTEST_INTERNAL = 1
        fixture_values = []
        for fx, fx_args in self.get_fixtures_except_for(f, kwds.keys()):
            assert inspect.isgeneratorfunction(
                fx
            ), f"function '{self.func.__name__}' cannot have async fixture '{fx.__name__}'."
            ctxmgr_func = contextlib.contextmanager(fx)
            context_mgr = self.run_recursively(ctxmgr_func, contextmgrstack, *fx_args)
            fixture_values.append(contextmgrstack.enter_context(context_mgr))
        return f(*fixture_values, *args, **kwds)

    # compare ^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v^v and test

    async def async_run_recursively(self, f, contextmgrstack, *args, **kwds):
        SELFTEST_INTERNAL = 1
        fixture_values = []
        for fx, fx_args in self.get_fixtures_except_for(f, kwds.keys()):
            ctxmgr_func = contextlib.asynccontextmanager(
                ensure_async_generator_func(fx)
            )
            context_mgr = await self.async_run_recursively(
                ctxmgr_func, contextmgrstack, *fx_args
            )
            fixture_values.append(
                await contextmgrstack.enter_async_context(context_mgr)
            )
        return f(*fixture_values, *args, **kwds)

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    async def async_run_with_fixtures(self, *args, **kwargs):
        SELFTEST_INTERNAL = 1
        async with contextlib.AsyncExitStack() as contextmgrstack:
            result = await self.async_run_recursively(
                self.func, contextmgrstack, *args, **kwargs
            )
            assert inspect.iscoroutine(result)
            loop = asyncio.get_running_loop()
            loop.set_exception_handler(asyncio_filtering_exception_handler)
            await asyncio.wait_for(result, timeout=None)  # timeout via async test


fixtures_hook = _Fixtures


# some general purpose standard fixtures


def tmp_path(name=None):
    with tempfile.TemporaryDirectory() as p:
        p = pathlib.Path(p)
        if name:
            yield p / name
        else:
            yield p


def _capture(name):
    """captures output from child processes as well"""
    org_stream = getattr(sys, name)
    org_fd = org_stream.fileno()
    org_fd_backup = os.dup(org_fd)
    replacement = tempfile.TemporaryFile(mode="w+t", buffering=1)
    os.dup2(replacement.fileno(), org_fd)
    setattr(sys, name, replacement)

    def getvalue():
        replacement.flush()
        replacement.seek(0)
        return replacement.read()

    replacement.getvalue = getvalue
    try:
        yield replacement
    finally:
        os.dup2(org_fd_backup, org_fd)
        setattr(sys, name, org_stream)


def stdout():
    yield from _capture("stdout")


def stderr():
    yield from _capture("stderr")


def raises(exception=Exception, message=None):
    try:
        yield
    except exception as e:
        if message and message != str(e):
            raise AssertionError(
                f"should raise {exception.__name__} with message '{message}'"
            ) from e
    except BaseException as e:
        raise AssertionError(
            f"should raise {exception.__name__} but raised {type(e).__name__}"
        ).with_traceback(e.__traceback__) from e
    else:
        e = AssertionError(f"should raise {exception.__name__}")
        e.__suppress_context__ = True
        raise e


def guard():
    keep_sys_path = sys.path.copy()
    keep_meta_path = sys.meta_path.copy()
    keep_modules = sys.modules.copy()
    yield
    assert (
        keep_sys_path == sys.path
    ), f"sys.path contaminated: {set(keep_sys_path) ^ set(sys.path)}"
    assert (
        keep_meta_path == sys.meta_path
    ), f"sys.meta_path contaminated: {set(keep_meta_path) ^ set(sys.meta_path)}"
    assert (
        keep_modules == sys.modules
    ), f"sys.modules contaminated: {keep_modules.keys() ^ sys.modules.keys()}"


def argv():
    orgv = sys.argv[:]
    del sys.argv[1:]
    yield sys.argv
    sys.argv = orgv


def stdin():
    try:
        sys.stdin = io.StringIO()
        yield sys.stdin
    finally:
        sys.stdin = sys.__stdin__


std_fixtures = {fx.__name__: fx for fx in [tmp_path, stdout, stderr, stdin, raises, guard, argv]}


def capture_stdout_child_processes_spawn_f():
    from selftest import get_tester
    self_test = get_tester(__name__)

    @self_test(subprocess=True)
    def in_child():
        print("hier ben ik")
        assert 1 == 1


def fixtures_test(self_test):
    from .binder import binder_hook

    self_test = self_test.getChild(
        hooks=(fixtures_hook, binder_hook), fixtures=std_fixtures
    )

    trace = []

    @self_test.fixture
    def fx_a():
        trace.append("A start")
        yield 67
        trace.append("A end")

    @self_test.fixture
    def fx_b(fx_a):
        trace.append("B start")
        yield 74
        trace.append("B end")

    def test_a(fx_a, fx_b, a, b=10):
        assert 67 == fx_a
        assert 74 == fx_b
        assert 9 == a
        assert 11 == b
        trace.append("test_a done")

    fixture_lifetime = []

    @self_test.fixture
    def fixture_A():
        fixture_lifetime.append("A-live")
        yield 42
        fixture_lifetime.append("A-close")

    @self_test
    def with_one_fixture(fixture_A):
        assert 42 == fixture_A, fixture_A

    @self_test.fixture
    def fixture_B(fixture_A):
        fixture_lifetime.append("B-live")
        yield fixture_A * 2
        fixture_lifetime.append("B-close")

    @self_test.fixture
    def fixture_C(fixture_B):
        fixture_lifetime.append("C-live")
        yield fixture_B * 3
        fixture_lifetime.append("C-close")

    @self_test
    def nested_fixture(fixture_B):
        assert 84 == fixture_B, fixture_B

    class lifetime:
        del fixture_lifetime[:]

        @self_test
        def more_nested_fixtures(fixture_C):
            assert 252 == fixture_C, fixture_C
            assert ["A-live", "B-live", "C-live"] == fixture_lifetime, fixture_lifetime

        @self_test
        def fixtures_livetime():
            assert [
                "A-live",
                "B-live",
                "C-live",
                "C-close",
                "B-close",
                "A-close",
            ] == fixture_lifetime, fixture_lifetime

    @self_test
    def stdout_capture():
        name = "Erik"
        msgs = []
        sys_stdout = sys.stdout
        sys_stderr = sys.stderr

        @self_test
        def capture_all(stdout, stderr):
            print(f"Hello {name}!", file=sys.stdout)
            print(f"Bye {name}!", file=sys.stderr)
            msgs.extend([stdout.getvalue(), stderr.getvalue()])
            self_test.ne(sys_stdout, sys.stdout)
            self_test.ne(sys_stderr, sys.stderr)

        self_test.eq("Hello Erik!\n", msgs[0])
        self_test.eq("Bye Erik!\n", msgs[1])
        self_test.eq(sys_stdout, sys.stdout)
        self_test.eq(sys_stderr, sys.stderr)

    @self_test
    def capture_stdout_child_processes(stdout):
        import multiprocessing
        def f():
            @self_test
            def in_child():
                print("hier ben ik")
                assert 1 == 1
        target = f

        if multiprocessing.get_start_method() != 'fork':
            target = capture_stdout_child_processes_spawn_f


        p = multiprocessing.Process(target=target)
        p.start()
        p.join(1)
        s = stdout.getvalue()
        assert "hier ben ik\n" in s, s

    @self_test
    def argv_fixture():
        org_arvg = sys.argv[:]
        with self_test.argv as a:
            a += ['new_arg']
            self_test.eq('new_arg', sys.argv[-1])
            self_test.eq(org_arvg + ['new_arg'], sys.argv)
        self_test.eq(org_arvg, sys.argv)

    @self_test
    def stdin_fixture():
        with self_test.stdin as i:
            i.write('abc\n')
            i.seek(0)
            self_test.eq('abc', input())

    @self_test
    def test_fixtures():
        with self_test.stdout as s:
            print("hello!")
            assert "hello!\n" == s.getvalue()
        keep = []
        with self_test.tmp_path as p:
            keep.append(p)
            (p / "f").write_text("contents")
        assert not keep[0].exists()

    # with self_test.<fixture> does not need @test to run 'in'
    with self_test.tmp_path as p:
        assert p.exists()
    assert not p.exists()

    @self_test
    def override_fixtures_in_new_context():
        with self_test.child() as t:

            @t.fixture
            def temporary_fixture():
                yield "tmp one"

            with t.temporary_fixture as tf1:
                assert "tmp one" == tf1, tf1
            with self_test.child() as child:

                @child.fixture
                def temporary_fixture():
                    yield "tmp two"

                with child.temporary_fixture as tf2:
                    assert "tmp two" == tf2
            with t.temporary_fixture as tf1:
                assert "tmp one" == tf1, tf1
        try:
            self_test.temporary_fixture
        except AttributeError:
            pass

    @self_test.fixture
    def area(r, d=1):
        import math

        yield round(math.pi * r * r, d)

    @self_test
    def fixtures_with_1_arg(area: 3):
        self_test.eq(28.3, area)

    @self_test
    def fixtures_with_2_args(area: (3, 0)):
        self_test.eq(28.0, area)

    @self_test.fixture
    def answer():
        yield 42

    @self_test.fixture
    def combine(a, area: 2, answer):
        yield a * area * answer

    @self_test
    def fixtures_with_combined_args(combine: 3):
        self_test.eq(1587.6, combine)

    @self_test
    def test_calls_other_test_with_fixture():
        @self_test(keep=True)
        def test_a(fixture_A):
            assert 42 == fixture_A
            return True

        @self_test
        def test_b():
            assert test_a(fixture_A=42)

    @self_test
    def test_calls_other_test_with_fixture_and_more_args():
        @self_test(keep=True, run=False)
        def test_a(fixture_A, value):
            assert 42 == fixture_A
            assert 16 == value
            return True

        @self_test
        def test_b(fixture_A):
            assert test_a(fixture_A=42, value=16)

        v = 45

        @self_test.fixture
        def f_A():
            yield v

        @self_test
        def fixtures_can_also_see_attrs_from_classed_being_defined(f_A):
            assert v == f_A, f_A

    @self_test.fixture
    def fixture_D(fixture_A, a=10):
        yield a

    @self_test
    def use_fixtures_as_context():
        with self_test.fixture_A as a:
            assert 42 == a
        with self_test.fixture_B as b:  # or pass parameter/fixture ourselves?
            assert 84 == b
        with self_test.fixture_C as c:
            assert 252 == c
        # it is possible to supply additional args when used as context
        with self_test.fixture_D as d:  # only fixture as arg
            assert 10 == d
        with self_test.fixture_D(16) as d:  # fixture arg + addtional arg
            assert 16 == d

    @self_test
    def bind_test_functions_to_their_fixtures():
        @self_test.fixture
        def my_fix():
            yield 34

        @self_test(run=False, keep=True)
        def override_fixture_binding_with_kwarg(my_fix):
            assert 34 != my_fix, "should be 34"
            assert 56 == my_fix
            return my_fix

        v = override_fixture_binding_with_kwarg(my_fix=56)  # fixture binding overridden
        assert v == 56

        try:
            self_test(override_fixture_binding_with_kwarg)
        except AssertionError as e:
            self_test.eq("should be 34", str(e))

        @self_test(keep=True)
        def bound_fixture_1(my_fix):
            assert 34 == my_fix, my_fix
            return my_fix

        # general way to rerun a test with other fixtures:
        with self_test.child() as t:

            @t.fixture
            def my_fix():
                yield 89

            try:
                t(bound_fixture_1)
            except AssertionError as e:
                assert "89" == str(e)
        with self_test.my_fix as x:
            assert 34 == x, x  # old fixture back

        @self_test.fixture
        def my_fix():  # redefine fixture purposely to test time of binding
            yield 78

        class A:
            a = 34

            @self_test(keep=True, run=False)
            def bound_fixture_acces_class_locals(my_fix):
                assert 78 == my_fix
                assert 34 == a
                return a

            # TODO function should have an extended closure, but when kept
            #      the original function is kept.
            # assert 34 == bound_fixture_acces_class_locals(78)

        trace = []

        @self_test.fixture
        def enumerate():
            trace.append("S")
            yield 1
            trace.append("E")

        @self_test(keep=True, run=False)
        def rebind_on_every_call(enumerate):
            return True

        assert [] == trace
        self_test(rebind_on_every_call)
        assert ["S", "E"] == trace
        self_test(rebind_on_every_call)
        assert ["S", "E", "S", "E"] == trace

    # below is an extra test to assure fixtures work with nested async funcs
    from .asyncer import async_hook

    with self_test.child(hooks=[async_hook]) as atest:
        done = [False, False, False]

        @atest
        async def this_is_an_async_test():
            done[0] = True
            """ A decorator is always called synchronously, so it can't call the async test
                because an event loop is already running. Solution is to start a new loop."""

            @atest
            async def this_is_a_nested_async_test():
                done[1] = True

                @atest
                async def this_is_a_doubly_nested_async_test():
                    done[2] = True

            atest.all(done)

    @self_test
    def assert_raises():
        with self_test.raises:
            raise Exception
        try:
            with self_test.raises:
                pass
        except AssertionError as e:
            assert "should raise Exception" == str(e), e

    @self_test
    def assert_raises_specific_exception():
        with self_test.raises(KeyError):
            raise KeyError
        try:
            with self_test.raises(KeyError):
                raise RuntimeError("oops")
        except AssertionError as e:
            assert "should raise KeyError but raised RuntimeError" == str(e), str(e)
        try:
            with self_test.raises(KeyError):
                pass
        except AssertionError as e:
            assert "should raise KeyError" == str(e), e

    @self_test
    def assert_raises_specific_message():
        with self_test.raises(RuntimeError, "hey man!"):
            raise RuntimeError("hey man!")
        try:
            with self_test.raises(RuntimeError, "hey woman!"):
                raise RuntimeError("hey man!")
        except AssertionError as e:
            assert "should raise RuntimeError with message 'hey woman!'" == str(e)

    @self_test
    def assert_raises_as_fixture(raises: KeyError):
        {}[0]

    @self_test(my_option=56)
    def access_test(test):
        """implicit fixture test"""
        assert test.option_get("my_option") == 56

    @self_test
    def guard_sys_path(tmp_path):
        try:
            with self_test.guard as p:
                sys.path.append(tmp_path)
            self.fail()
        except AssertionError as e:
            self_test.eq(f"sys.path contaminated: {{{tmp_path!r}}}", e.args[0])
        p = sys.path.pop()
        assert p == tmp_path

    @self_test
    def guard_meta_path():
        try:
            with self_test.guard as p:
                sys.meta_path.append("does not work")
            self.fail()
        except AssertionError as e:
            self_test.eq(f"sys.meta_path contaminated: {{'does not work'}}", e.args[0])
        p = sys.meta_path.pop()
        assert p == "does not work"

    @self_test
    def guard_modules():
        try:
            with self_test.guard as p:
                sys.modules["this_is_a_module"] = "not a module"
            self.fail()
        except AssertionError as e:
            self_test.eq(f"sys.modules contaminated: {{'this_is_a_module'}}", e.args[0])
        p = sys.modules.pop("this_is_a_module")
        assert p == "not a module"




## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

from .utils import extend_closure  # redefine placeholder


class _Binder:
    def __call__(self, runner, func):
        if runner.option_get("bind", False):
            return extend_closure(func)
        return func


binder_hook = _Binder()


def binder_test(self_test):
    self_test2 = self_test.getChild(hooks=(binder_hook,))

    class binding_context:
        a = 42

        @self_test2(keep=True, bind=True)
        def one_test():
            assert a == 42

        a = 43
        # one_test()

    @self_test2
    def access_closure_from_enclosing_def():
        a = 46  # accessing this in another function makes it a 'freevar', which needs a closure

        @self_test2
        def access_a():
            assert 46 == a

    f = 16

    @self_test2
    def dont_confuse_app_vars_with_internal_vars():
        assert 16 == f



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import logging

""" NB
    Tests are not assigned levels, Runners are. The level of the Runner
    determines the level of the tests run with it. So a single Runner
    will run all tests, since the level of the tests is by definition
    equal to the level of the runner.

    Filtering only becomes effective between Runners of different levels.
    If a child has level >= its parent level, all tests will run. If a child
    has a level < the level of its parent, it wil not run.

    The catch is that an individual test marked with e.g. @test.unit will
    run in a one-off temporary anonymous child of 'test' with the level
    UNIT, in this case.

    This relation extends out to arbitrarily deep nested Runners. A test
    only runs when its level is >= all parent levels. That is, any parent
    can raise the threshold in order to exclude tests.
"""

CRITICAL = 50
UNIT = 40
INTEGRATION = 30
PERFORMANCE = 20
NOTSET = 0


levels = {
    "CRITICAL": CRITICAL,
    "UNIT": UNIT,
    "INTEGRATION": INTEGRATION,
    "PERFORMANCE": PERFORMANCE,
    "NOTSET": NOTSET,
    CRITICAL: "CRITICAL",
    UNIT: "UNIT",
    INTEGRATION: "INTEGRATION",
    PERFORMANCE: "PERFORMANCE",
    NOTSET: "NOTSET",
}

DEFAULT_LEVEL = UNIT
DEFAULT_THRESHOLD = INTEGRATION


def numeric_level(l):
    return levels[l.upper()] if isinstance(l, str) else l


class _Levels:
    def __call__(self, tester, func):
        if tester.level >= tester.threshold:
            return func

    def lookup(self, tester, name):
        if name == "level":
            return numeric_level(tester.option_get(name, DEFAULT_LEVEL))
        if name == "threshold":
            return max(
                (numeric_level(l) for l in tester.option_enumerate("threshold")),
                default=DEFAULT_THRESHOLD,
            )
        if (level := levels.get(name.upper())) is not None:
            return tester(level=level)
        raise AttributeError

    def logrecord(self, tester, func, record):
        level = tester.level
        strlevel = levels[level]
        record.testlevel = level
        record.testlevelname = strlevel
        record.msg = f"{strlevel}:{record.msg}"
        return record


levels_hook = _Levels()


def levels_test(self_test):
    def run_various_tests(test):
        runs = [0, 0, 0, 0, 0, 0]

        @test.critical
        def a_critial_test():
            runs[0] = 1

        @test.unit
        def a_unit_test():
            runs[1] = 1

        @test.integration
        def a_integration_test():
            runs[2] = 1

        @test.performance
        def a_performance_test():
            runs[3] = 1

        @test.notset
        def a_notset_test():
            runs[4] = 1

        @test
        def a_default_test():
            runs[5] = 1

        return runs

    @self_test
    def probe_various_levels_and_thresholds():
        with self_test.child(
            hooks=(levels_hook,), level="unit", threshold="critical"
        ) as test:
            r = run_various_tests(test)
            assert [1, 0, 0, 0, 0, 0] == r, r

        with self_test.child(
            hooks=(levels_hook,), level="notset", threshold="unit"
        ) as test:
            r = run_various_tests(test)
            assert [1, 1, 0, 0, 0, 0] == r, r

        with self_test.child(
            hooks=(levels_hook,), level="performance", threshold="integration"
        ) as test:
            r = run_various_tests(test)
            assert [1, 1, 1, 0, 0, 0] == r, r

        with self_test.child(
            hooks=(levels_hook,), level="integration", threshold="performance"
        ) as test:
            r = run_various_tests(test)
            assert [1, 1, 1, 1, 0, 1] == r, r

        with self_test.child(
            hooks=(levels_hook,), level="critical", threshold="notset"
        ) as test:
            r = run_various_tests(test)
            assert [1, 1, 1, 1, 1, 1] == r, r

        with self_test.child(
            hooks=(levels_hook,), level="unit", threshold="integration"
        ) as test:
            r = run_various_tests(test)
            assert [1, 1, 1, 0, 0, 1] == r, r
            with test.child(
                level="performance", threshold="notset"
            ) as test2:  # threshold ignored
                r = run_various_tests(test2)
                assert [1, 1, 1, 0, 0, 0] == r, r
            with test.child(
                level="integration", threshold="performance"
            ) as test2:  # threshold ignored
                r = run_various_tests(test2)
                assert [1, 1, 1, 0, 0, 1] == r, r



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##


def any(_):  # name is included in diffs
    return True


class _Any:
    def __init__(self, f=any):
        self.f = f

    def __eq__(self, x):
        return bool(self.f(x))

    def __call__(self, f):
        return _Any(f)

    def __repr__(self):
        return self.f.__name__ + "(...)" if self.f else "*"


class _Wildcard:
    def __call__(self, tester, func):
        return func

    def lookup(self, tester, name):
        if name == "any":
            return _Any()
        raise AttributeError


wildcard_hook = _Wildcard()


def wildcard_test(self_test):
    with self_test.child(hooks=(wildcard_hook,)) as test:

        @test
        def wildcard_matching():
            test.eq([test.any, 42], [16, 42])
            test.eq([test.any(lambda x: x in [1, 2, 3]), 78], [2, 78])
            test.ne([test.any(lambda x: x in [1, 2, 3]), 78], [4, 78])



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2021-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##


"""
    The structure of selftests for bootstrapping.
    ---------------------------------------------
    * tester.py contains the main Tester without any hooks, this module tests itself using a
      separate Tester called self_test. We reuse self_test to incrementally run the tests for
      the hooks.
    * self_test contains one hook: operator, in order to make testing easier. However, a
      small mistake in operator might cause all tests to fail since Tester and operator
      mutually depend on each other.
    * After all hooks have been tested, we assemble the final root Tester and tests if
      all hooks work properly.
    * Run tests for selftest itself by:
       $ python3 -c "import selftest" selftest.selftest
    * Integration tests are run with a partially initialized __init__.py, but it works
    * Finally we test the setup

"""


import os


from .tester import Tester, self_test


@self_test
def assert_stats():
    assert {"found": 21, "run": 20} == self_test.stats, self_test.stats


from .levels import levels_hook, levels_test

levels_test(self_test)

from .prrint import prrint_test

prrint_test(self_test)

from .filter import filter_hook, filter_test

filter_test(self_test)

from .wildcard import wildcard_hook, wildcard_test

wildcard_test(self_test)

from .binder import binder_hook, binder_test

binder_test(self_test)

from .operators import operators_hook, operators_test

operators_test(self_test)

from .fixtures import fixtures_hook, fixtures_test, std_fixtures

fixtures_test(self_test)

from .asyncer import async_hook, async_test

async_test(self_test)

from .diffs import diff_hook, diff_test

diff_test(self_test)


@self_test
def check_stats():
    self_test.eq({"found": 157, "run": 131}, self_test.stats)


def assemble_root_runner(**options):
    return Tester(
        # order of hook matters, processed from right to left
        hooks=[
            operators_hook,
            async_hook,
            fixtures_hook,
            diff_hook,
            levels_hook,
            wildcard_hook,
            binder_hook,
            filter_hook,
        ],
        fixtures=std_fixtures,
        **options,
    )


def root_tester_assembly_test(test):
    """only test if the root tester is assembled with all hooks"""

    N = [0]

    # operators_hook
    test.eq(1, 1)
    test.isinstance({}, dict)
    test.endswith("aap", "ap")

    # fixtures hook
    @test.fixture
    def forty_two():
        yield 42

    @test
    def is_forty_two(forty_two):
        assert forty_two == 42
        N[0] += 1

    with test.forty_two as contextmanager:
        assert contextmanager == 42

    # standard fixtures
    with test.tmp_path as p:
        test.truth(p.exists())
    with test.stdout as e:
        print("one")
        assert e.getvalue() == "one\n"
    with test.stderr as e:
        import sys

        print("two", file=sys.stderr)
        assert e.getvalue() == "two\n"

    # binding hook
    class A:
        a = 42

        @test(bind=True)
        def bind():
            assert a == 42
            N[0] += 1

    # wildcard hook
    test.eq(test.any(int), 42)

    # levels hook
    from .levels import UNIT

    assert test.level == 40, test.level
    assert test.threshold == 30, test.threshold
    with test.child(threshold=UNIT) as tst:

        @tst.performance
        def performance_test():
            assert "not" == "executed"

        @tst.critical
        def critical_test():
            assert 1 == 1
            N[0] += 1

        assert {"found": 2, "run": 1} == tst.stats, tst.stats

    # async hook (elaborate on nested stuff)
    @test.fixture
    async def nine():
        yield [
            "borg 1",
            "borg 2",
            "borg 3",
            "borg 4",
            "borg 5",
            "borg 6",
            "Annika Hansen",
            "borg 8",
            "borg 9",
        ]

    @test.fixture
    async def seven_of_nine(nine):
        yield nine[7 - 1]

    @test
    async def the_9(nine):
        assert len(nine) == 9
        N[0] += 1

    @test
    async def is_seven_of_nine(seven_of_nine):
        assert seven_of_nine == "Annika Hansen"
        N[0] += 1

    assert N[0] == 5, N
    assert dict(found=6, run=5) == test.stats, test.stats

    # diff hook
    try:
        test.eq(1, 2, diff=test.diff)
    except AssertionError as e:
        assert str(e) == "\n- 1\n+ 2"
    try:
        test.eq(1, 2, diff=test.diff2)
    except AssertionError as e:
        assert str(e) == "\n- 1\n+ 2"

    # filter hook
    @test(filter="aa")
    def moon():
        test.fail()

    r = [0]

    @test(filter="aa")
    def maan():
        r[0] = 1

    assert r == [1]


testers = {}  # initial, for more testing


def basic_config(**options):
    # raise Exception
    CR = '\n'
    assert None not in testers, f"Root Tester {testers[None]}{CR} already created."
    testers[None] = assemble_root_runner(**options)


def get_tester(name=None):
    if None not in testers:
        testers[None] = assemble_root_runner()
    if name in testers:
        return testers[name]
    tester = testers[None]
    for namepart in name.split("."):
        tester = tester.getChild(namepart)
        testers[tester._name] = tester
    return tester


@self_test
def set_root_opts():
    basic_config(filter="aap")
    root = get_tester()
    self_test.eq(
        {"filter", "fixtures", "hooks", "keep", "run", "subprocess"},
        set(root._options.keys()),
    )
    self_test.eq("aap", root._options["filter"])


@self_test
def get_root_tester():
    root = get_tester()
    assert isinstance(root, Tester)
    root1 = get_tester()
    assert root1 is root


@self_test
def get_sub_tester():
    root = get_tester()
    mymodule = get_tester("my.module")
    assert mymodule._name == "my.module"
    my = get_tester("my")
    assert my._name == "my"
    assert my._parent is root
    assert mymodule._parent is my
    mymodule1 = get_tester("my.module")
    assert mymodule1 is mymodule


testers.clear()


@self_test
def run_integration_tests():
    root_tester_assembly_test(assemble_root_runner())
    from .integrationtests import integration_test

    integration_test(assemble_root_runner().integration)


testers.clear()


@self_test
def setup_correct():
    import tempfile
    import pathlib

    with tempfile.TemporaryDirectory() as p:
        tmp = pathlib.Path(p)
        selftest_dev_dir = pathlib.Path(__file__).parent.resolve().parent
        if not (selftest_dev_dir / "bin/selftest").exists():
            # Not dev dir
            return
        import subprocess

        version_process = subprocess.run(
            ["python3", "setup.py", "--version"],
            capture_output=True,
            text=True,
            cwd=str(selftest_dev_dir),
        )
        version = version_process.stdout.strip()
        result = subprocess.run(
            ["python3", "setup.py", "sdist", "--dist-dir", str(tmp)],
            capture_output=True,
            cwd=str(selftest_dev_dir),
        )

        from tarfile import open

        tf = open(name=tmp / f"selftest-{version}.tar.gz", mode="r:gz")
        self_test.eq(
            [
                f"selftest-{version}",
                f"selftest-{version}/LICENSE",
                f"selftest-{version}/MANIFEST.in",
                f"selftest-{version}/PKG-INFO",
                f"selftest-{version}/README.rst",
                f"selftest-{version}/bin",
                f"selftest-{version}/bin/selftest",
                f"selftest-{version}/selftest",
                f"selftest-{version}/selftest.egg-info",
                f"selftest-{version}/selftest.egg-info/PKG-INFO",
                f"selftest-{version}/selftest.egg-info/SOURCES.txt",
                f"selftest-{version}/selftest.egg-info/dependency_links.txt",
                f"selftest-{version}/selftest.egg-info/top_level.txt",
                f"selftest-{version}/selftest/__init__.py",
                f"selftest-{version}/selftest/__main__.py",
                f"selftest-{version}/selftest/asyncer.py",
                f"selftest-{version}/selftest/binder.py",
                f"selftest-{version}/selftest/diffs.py",
                f"selftest-{version}/selftest/filter.py",
                f"selftest-{version}/selftest/fixtures.py",
                f"selftest-{version}/selftest/integrationtests.py",
                f"selftest-{version}/selftest/levels.py",
                f"selftest-{version}/selftest/mocks.py",
                f"selftest-{version}/selftest/operators.py",
                f"selftest-{version}/selftest/prrint.py",
                f"selftest-{version}/selftest/tester.py",
                f"selftest-{version}/selftest/tests",
                f"selftest-{version}/selftest/tests/__init__.py",
                f"selftest-{version}/selftest/tests/sub_module_fail.py",
                f"selftest-{version}/selftest/tests/sub_module_ok.py",
                f"selftest-{version}/selftest/tests/temporary_class_namespace.py",
                f"selftest-{version}/selftest/tests/tryout.py",
                f"selftest-{version}/selftest/tests/tryout2.py",
                f"selftest-{version}/selftest/utils.py",
                f"selftest-{version}/selftest/wildcard.py",
                f"selftest-{version}/setup.cfg",
                f"selftest-{version}/setup.py",
            ],
            sorted(tf.getnames()),
            diff=lambda a, b: set(a).symmetric_difference(set(b)),
        )
        tf.close()


"""
We put these last, as printing any debug/trace messages anywhere in the code causes this
to fail.
"""

with self_test.child(hooks=[fixtures_hook], fixtures=std_fixtures) as self_test2:

    @self_test2
    def main_without_args(stdout):
        os.system("PYTHONPATH=. python3 selftest")
        assert "Usage: selftest [options] module" in stdout.getvalue()

    @self_test2
    def main_without_help(stdout):
        os.system("PYTHONPATH=. python3 selftest --help")
        s = stdout.getvalue()
        assert "Usage: selftest [options] module" in s
        assert "-h, --help            show this help message and exit" in s
        assert "-f FILTER, --filter=FILTER" in s
        assert "only run tests whose qualified name contains FILTER" in s
        assert "-t THRESHOLD, --threshold=THRESHOLD" in s
        assert "only run tests whose level is >= THRESHOLD" in s

    @self_test2
    def main_test(stderr, stdout):
        import os

        os.system("PYTHONPATH=. python3 selftest selftest/tests/tryout.py")
        e = stderr.getvalue()
        s = stdout.getvalue()
        assert "" == s, s
        loglines = [l for l in e.splitlines() if not '^^^^' in l]
        assert "importing selftest.tests.tryout" in loglines[0], loglines[0]
        assert loglines[1].startswith(
            "\033[1mTEST\033[0m:\033[1mUNIT:selftest.tests.tryout.one_simple_test\033[0m:"
        ), loglines[1]
        assert loglines[1].endswith("/selftest/selftest/tests/tryout.py:29"), loglines[ 1 ]
        assert loglines[2].startswith(
            "\033[1mTEST\033[0m:\033[1mINTEGRATION:selftest.tests.tryout.one_more_test\033[0m:"
        ), loglines[2]
        assert loglines[2].endswith("/selftest/selftest/tests/tryout.py:34"), loglines[ 2 ]
        assert " 31  \t    test.eq(1, 1)" == loglines[3]
        assert " 32  \t" == loglines[4]
        assert " 33  \t" == loglines[5]
        assert " 34  \t@test.integration" == loglines[6]
        assert " 35  \tasync def one_more_test():" == loglines[7]
        assert ' 36  ->\t    assert 1 == 2, "one is not two"' == loglines[8]
        assert " 37  \t    test.eq(1, 2)" == loglines[9]
        assert "[EOF]" == loglines[10]
        assert "Traceback (most recent call last):" in loglines[11]
        # some stuff in between we can't get rid off
        #
        import sys
        assert "in <module>" in loglines[-5]
        assert "selftest/tests/tryout.py" in loglines[-5]
        if sys.version_info.minor < 11:
            assert "async def one_more_test():" in loglines[-4], '\n'.join(loglines[-5:])
        assert "in one_more_test" in loglines[-3]
        assert "selftest/tests/tryout.py" in loglines[-3]
        assert "assert 1 == 2" in loglines[-2]
        assert "AssertionError: one is not two" in loglines[-1], loglines[-1]
        expected_length = 21
        if sys.version_info.minor < 11:
            expected_length = 23
        assert expected_length == len(loglines), f'{len(loglines)} ' + "\n".join(repr(l) for l in loglines)

    # @self_test2
    def main_with_selftests(stdout, stderr):
        os.system("PYTHONPATH=. python3 selftest selftest.selftest")
        lns = stdout.getvalue().splitlines()
        assert ["Usage: selftest [options] module", ""] == lns, lns
        lns = stderr.getvalue().splitlines()
        assert len(lns) == 144, len(lns)  # number of logged tests, sort of

    def assert_output(stdout, stderr):
        o = stdout.getvalue()
        self_test2.eq("", o)
        lines = stderr.getvalue().splitlines()
        self_test2.startswith(
            lines[0],
            "\033[1mTEST\033[0m:\033[1mimporting selftest.tests.tryout2\033[0m:",
        )
        self_test2.startswith(
            lines[1],
            "\033[1mTEST\033[0m:\033[1mUNIT:selftest.tests.tryout2.one_simple_test\033[0m:",
        )
        self_test2.startswith(
            lines[2],
            "\033[1mTEST\033[0m:\033[1mINTEGRATION:selftest.tests.tryout2.one_integration_test\033[0m:",
        )
        self_test2.startswith(
            lines[3], "\033[1mTEST\033[0m:\033[1mstats: found: 3, run: 2\033[0m:"
        )

    @self_test2
    def main_via_bin_script_with_selftest_on_path(stdout, stderr):
        os.system(f"PATH=./bin:$PATH selftest selftest/tests/tryout2.py")
        assert_output(stdout, stderr)

    @self_test2
    def main_via_bin_script_in_cur_dir(stdout, stderr):
        os.system(f"(cd bin; ./selftest selftest/tests/tryout2.py)")
        assert_output(stdout, stderr)

    @self_test2
    def main_with_filter(stdout, stderr):
        os.system(
            "PYTHONPATH=. python3 selftest selftest/tests/tryout2.py --filter one_simple"
        )
        o = stdout.getvalue()
        assert "" == o, o
        e = stderr.getvalue()
        assert "one_simple_test" in e, e
        assert "one_integration_test" not in e, e
        assert "one_performance_test" not in e
        assert "found: 3, run: 1" in e, e

    @self_test2
    def main_with_level_unit(stdout, stderr):
        os.system(
            "PYTHONPATH=. python3 selftest selftest/tests/tryout2.py --threshold unit"
        )
        o = stdout.getvalue()
        assert "" == o, o
        e = stderr.getvalue()
        assert "one_simple_test" in e, e
        assert "one_integration_test" not in e, e
        assert "one_performance_test" not in e
        assert "found: 3, run: 1" in e, e

    @self_test2
    def main_with_level_integration(stdout, stderr):
        os.system(
            "PYTHONPATH=. python3 selftest selftest/tests/tryout2.py --threshold integration"
        )
        o = stdout.getvalue()
        assert "" == o, o
        e = stderr.getvalue()
        assert "one_simple_test" in e, e
        assert "one_integration_test" in e, e
        assert "one_performance_test" not in e
        assert "found: 3, run: 2" in e, e



#!/usr/bin/env python3
## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.rst").read_text(encoding="utf-8")

setup(
    name="selftest",
    description="Python In-Source Testing Library",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=find_packages(),
    author="Erik Groeneveld",
    author_email="erik@seecr.nl",
    maintainer="Thijs Janssen",
    maintainer_email="thijs@seecr.nl",
    url="https://github.com/ejgroene/selftest",
    scripts=["bin/selftest"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Testing :: Unit",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.9",
    setuptools_git_versioning={ "enabled": True },
    setup_requires=["setuptools-git-versioning>=2.0,<3"]
)



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##

import contextlib
import types  # for creating Function object with extended bindings
import inspect
import sys
import functools  # wrapping function in async generator


def iterate(f, v):
    while v:
        yield v
        v = f(v)


# get the type of Traceback objects
try:
    raise Exception
except Exception as e:
    TracebackType = type(e.__traceback__)
    del e


def _():
    yield


async def _a():
    yield


ContextManagerType = type(contextlib.contextmanager(_)())
AsyncContextManagerType = type(contextlib.asynccontextmanager(_a)())
del _, _a


def frame_to_traceback(tb_frame, tb_next=None):
    tb = TracebackType(tb_next, tb_frame, tb_frame.f_lasti, tb_frame.f_lineno)
    return create_traceback(tb_frame.f_back, tb) if tb_frame.f_back else tb


def bind_names(bindings, names, frame):
    """find names in locals on the stack and binds them"""
    # TODO restrict this a bit; now you can poke anywhere
    if not frame or not names:
        return bindings
    if not is_internal(frame):
        f_locals = frame.f_locals  # rather expensive
        rest = []
        for name in names:
            try:
                bindings[name] = f_locals[name]
            except KeyError:
                rest.append(name)
    else:
        rest = names
    return bind_names(bindings, rest, frame.f_back)


def extend_closure(func):
    """Binds the unbound vars in func to values found on the stack"""
    func2 = types.FunctionType(
        func.__code__,  # code
        bind_names(  # globals
            func.__globals__.copy(),
            func.__code__.co_names,  # names other than arguments and function locals
            inspect.currentframe().f_back,
        ),
        func.__name__,  # name
        func.__defaults__,  # default arguments
        func.__closure__,
    )  # closure/free vars
    func2.__annotations__ = func.__annotations__  # annotations not in __init__
    return func2


std_modules = sys.builtin_module_names + (
    "runpy",
    "concurrent.futures._base",
    "concurrent.futures.thread",
    "asyncio.runners",
    "asyncio.base_events",
    "asyncio.tasks",
    #'selftest',
    #'selftest.binder',
    "selftest.fixtures",
    "importlib",
)


def is_builtin(f):
    if m := inspect.getmodule(f.f_code):
        return m.__name__ in std_modules


def is_internal(frame):
    nm = frame.f_code.co_filename
    return (
        "SELFTEST_INTERNAL" in frame.f_code.co_varnames
        or "<frozen importlib" in nm
        or is_builtin(frame)
    )  # TODO testme


def ensure_async_generator_func(f):
    if inspect.isasyncgenfunction(f):
        return f
    if inspect.isgeneratorfunction(f):

        @functools.wraps(f)
        async def wrap(*a, **k):
            for v in f(*a, **k):
                yield v

        return wrap
    assert False, f"{f} cannot be a async generator."


# using import.import_module in asyncio somehow gives us the frozen tracebacks (which were
# removed in 2012, but yet showing up again in this case. Let's get rid of them.
def asyncio_filtering_exception_handler(loop, context):
    if "source_traceback" in context:
        context["source_traceback"] = [
            t for t in context["source_traceback"] if "<frozen " not in t.filename
        ]
    return loop.default_exception_handler(context)


class ArgsCollector:
    """collects args before calling a function: ArgsCollector(f, 1)(2)(3)() calls f(1,2,3)"""

    def __init__(self, f, *args, **kwds):
        self.func = f
        self.args = args
        self.kwds = kwds

    def __call__(self, *args, **kwds):
        if not args and not kwds:
            return self.func(*self.args, **self.kwds)
        self.args += args
        self.kwds.update(kwds)
        return self


class ArgsCollectingContextManager(ArgsCollector, ContextManagerType):
    """Context manager that accepts additional args everytime it is called.
    NB: Implementation closely tied to contexlib.py (self.gen)"""

    def __enter__(self):
        self.gen = self()
        return super().__enter__()


class ArgsCollectingAsyncContextManager(ArgsCollector, AsyncContextManagerType):
    async def __aenter__(self):
        self.gen = self()
        return await super().__aenter__()

    @property
    def __enter__(self):
        raise Exception(f"Use 'async with' for {self.func}.")


# @self_test #TODO
def extra_args_supplying_contextmanager():
    def f(a, b, c, *, d, e, f):
        yield a, b, c, d, e, f

    c = ArgsCollectingContextManager(f, "a", d="d")
    assert isinstance(c, contextlib.AbstractContextManager)
    c("b", e="e")
    with c("c", f="f") as v:
        assert ("a", "b", "c", "d", "e", "f") == v, v


def filter_traceback(root):
    # first prune the end of the list
    while root and is_internal(root.tb_frame):
        root = root.tb_next
    # then cut unwanted frames in between others
    tb = root
    while tb and tb.tb_next:
        if is_internal(tb.tb_next.tb_frame):
            tb.tb_next = tb.tb_next.tb_next
        else:
            tb = tb.tb_next
    return root


# @self_test2   #TODO
def trace_backfiltering():
    def eq(a, b):
        SELFTEST_INTERNAL = 1
        assert a == b

    def B():
        eq(1, 2)

    def B_in_betwixt():
        SELFTEST_INTERNAL = 1
        B()

    def A():
        B_in_betwixt()

    self_test2.contains(B_in_betwixt.__code__.co_varnames, "SELFTEST_INTERNAL")

    def test_names(*should):
        _, _, tb = sys.exc_info()
        tb = filter_traceback(tb)
        names = tuple(
            tb.tb_frame.f_code.co_name for tb in iterate(lambda tb: tb.tb_next, tb)
        )
        assert should == names, tuple(
            tb.tb_frame.f_code for tb in iterate(lambda tb: tb.tb_next, tb)
        )

    try:
        eq(1, 2)
    except AssertionError:
        test_names("trace_backfiltering")

    try:
        B()
    except AssertionError:
        test_names("trace_backfiltering", "B")

    try:
        B_in_betwixt()
    except AssertionError:
        test_names("trace_backfiltering", "B")

    try:
        A()
    except AssertionError:
        test_names("trace_backfiltering", "A", "B")

    def C():
        A()

    def D():
        SELFTEST_INTERNAL = 1
        C()

    def E():
        SELFTEST_INTERNAL = 1
        D()

    try:
        C()
    except AssertionError:
        test_names("trace_backfiltering", "C", "A", "B")

    try:
        D()
    except AssertionError:
        test_names("trace_backfiltering", "C", "A", "B")

    try:
        E()
    except AssertionError:
        test_names("trace_backfiltering", "C", "A", "B")

    try:

        @self_test2
        def test_for_real_with_Runner_involved():
            self_test2.eq(1, 2)

    except AssertionError:
        test_names("trace_backfiltering", "test_for_real_with_Runner_involved")



## begin license ##
#
# "selftest": a simpler test runner for python
#
# Copyright (C) 2022-2023 Seecr (Seek You Too B.V.) https://seecr.nl
#
# This file is part of "selftest"
#
# "selftest" is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "selftest" is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "selftest".  If not, see <http://www.gnu.org/licenses/>.
#
## end license ##


# This works:


def f():
    def g():
        print("g(1)")

    g()


f()


def f():
    def g():
        print("g(2)")

    def h():
        g()

    h()


f()


# This also works:


class f:
    def g():
        print("g(3)")

    g()


# but this does not:


class f:
    def g():
        print("g(4)")

    def h():
        g()  # "g is not defined"

    h()


# Explanation:
#
# 1.
# During the definition of def f, nothing is executed.
# During the definition of class f, the class body is executed.
# This is why f() is needed on line 12 and 25.
#
# 2.
# The def f() is completely defined before being executed: it's local namespace includes g.
# The class f() is incomplete when it is executed, g is only in a temporary namespace.
# Function h has its own namespace and has no access to class f's temporary namespace.
#
# 3.
# Calling g() directly does work, as this happens in the same temporary namespace.
